from .classic.aci import CraftaxClassicACI
from .full.aci import CraftaxACI
from .recording import EpisodeRecorder

from craftax.craftax_env import make_craftax_env_from_name

from .full.metadata import (
    CRAFTAX_FULL_ACHIEVEMENTS,
    CRAFTAX_FULL_ACTION_MAPPING,
)
from .full.state import render_craftax_text_custom
from .shared import (
    CraftaxBaseACI,
)

from .classic.metadata import (
    CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES,
    CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS,
    CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS,
    CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS,
    CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS,
)
from .shared import CraftaxState, mob_id_to_name

__all__ = [
    "CraftaxClassicACI",
    "CraftaxACI",
    "EpisodeRecorder",
    "make_craftax_env_from_name",
    "CRAFTAX_FULL_ACHIEVEMENTS",
    "CRAFTAX_FULL_ACTION_MAPPING",
    "render_craftax_text_custom",
    "CraftaxBaseACI",
    "CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES",
    "CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS",
    "CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS",
    "CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS",
    "CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS",
    "CraftaxState",
    "mob_id_to_name",
]



CRAFTAX_CLASSIC_ACHIEVEMENTS = {
    0: "Collect Wood",
    1: "Place Table",
    2: "Eat Cow",
    3: "Collect Sapling",
    4: "Collect Drink",
    5: "Make Wood Pickaxe",
    6: "Make Wood Sword",
    7: "Place Plant",
    8: "Defeat Zombie",
    9: "Collect Stone",
    10: "Place Stone",
    11: "Eat Plant",
    12: "Defeat Skeleton",
    13: "Make Stone Pickaxe",
    14: "Make Stone Sword",
    15: "Wake Up",
    16: "Place Furnace",
    17: "Collect Coal",
    18: "Collect Iron",
    19: "Collect Diamond",
    20: "Make Iron Pickaxe",
    21: "Make Iron Sword",
}
CRAFTAX_CLASSIC_ACTION_MAPPING = {
    "noop": 0,
    "left": 1,
    "right": 2,
    "up": 3,
    "down": 4,
    "do": 5,
    "sleep": 6,
    "place_stone": 7,
    "place_table": 8,
    "place_furnace": 9,
    "place_plant": 10,
    "make_wood_pickaxe": 11,
    "make_stone_pickaxe": 12,
    "make_iron_pickaxe": 13,
    "make_wood_sword": 14,
    "make_stone_sword": 15,
    "make_iron_sword": 16,
}

CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES = [
    "grass",
    "sand",
    "path",
]
CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS = [
    "water",
    "stone",
    "tree",
    "wood",
    "plant",
]
CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS = []
CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS = [
    "coal",
    "iron",
    "diamond",
    "crafting_table",
    "furnace",
    "lava",
    "ripe_plant",
]
CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS = ["Skeleton", "Zombie", "Cow", "Arrow"]



from craftax.craftax.constants import *
from craftax.craftax_env import make_craftax_env_from_name

from craftaxlm.full.metadata import (
    CRAFTAX_FULL_ACHIEVEMENTS,
    CRAFTAX_FULL_ACTION_MAPPING,
)
from craftaxlm.full.state import render_craftax_text_custom
from craftaxlm.shared import (
    CraftaxBaseACI,
)


class CraftaxACI(CraftaxBaseACI):
    def make_env(self):
        raise NotImplementedError("Update this with the same debug applied to Classic")
        return make_craftax_env_from_name("Craftax-Symbolic-v1", auto_reset=True)

    def create_starting_obs(self):
        return {
            "state": render_craftax_text_custom(self.state).render_to_text_simple(
                verbose=self.verbose
            )
        }

    def map_action_string_to_int(self, action_string: str) -> int:
        return CRAFTAX_FULL_ACTION_MAPPING.get(action_string.lower(), 0)

    def get_achievements(self, state):
        return {
            "achievements": {
                k: state.achievements[i] for i, k in CRAFTAX_FULL_ACHIEVEMENTS.items()
            }
        }

    def create_step_info(self, state, reward, done):
        return {
            "state": render_craftax_text_custom(state).render_to_text_simple(
                verbose=self.verbose
            ),
            "reward": float(reward),
            "done": bool(done),
        }


if __name__ == "__main__":
    craftax_aci = CraftaxACI()
    action = 0
    import time

    ts = []
    for i in range(20):
        t0 = time.time()
        step_info = craftax_aci._step(action)
        ts.append(time.time() - t0)
        print("Took", ts[-1])
        if step_info["done"]:
            break
    print("Mean time", sum(ts) / len(ts))
    print("Q3 time", sorted(ts)[int(len(ts) * 0.75)])
    print("P90 time", sorted(ts)[int(len(ts) * 0.90)])
    print("P99 time", sorted(ts)[int(len(ts) * 0.99)])



CRAFTAX_FULL_ACTION_MAPPING = {
    "noop": 0,
    "left": 1,
    "right": 2,
    "up": 3,
    "down": 4,
    "do": 5,
    "sleep": 6,
    "place_stone": 7,
    "place_table": 8,
    "place_furnace": 9,
    "place_plant": 10,
    "make_wood_pickaxe": 11,
    "make_stone_pickaxe": 12,
    "make_iron_pickaxe": 13,
    "make_wood_sword": 14,
    "make_stone_sword": 15,
    "make_iron_sword": 16,
    "rest": 17,
    "descend": 18,
    "ascend": 19,
    "make_diamond_pickaxe": 20,
    "make_diamond_sword": 21,
    "make_iron_armour": 22,
    "make_diamond_armour": 23,
    "shoot_arrow": 24,
    "make_arrow": 25,
    "cast_fireball": 26,
    "cast_iceball": 27,
    "place_torch": 28,
    "drink_potion_red": 29,
    "drink_potion_green": 30,
    "drink_potion_blue": 31,
    "drink_potion_pink": 32,
    "drink_potion_cyan": 33,
    "drink_potion_yellow": 34,
    "read_book": 35,
    "enchant_sword": 36,
    "enchant_armour": 37,
    "make_torch": 38,
    "level_up_dexterity": 39,
    "level_up_strength": 40,
    "level_up_intelligence": 41,
    "enchant_bow": 42,
}
CRAFTAX_FULL_ACHIEVEMENTS = {
    0: "Collect Wood",
    1: "Place Table",
    2: "Eat Cow",
    3: "Collect Sapling",
    4: "Collect Drink",
    5: "Make Wood Pickaxe",
    6: "Make Wood Sword",
    7: "Place Plant",
    8: "Defeat Zombie",
    9: "Collect Stone",
    10: "Place Stone",
    11: "Eat Plant",
    12: "Defeat Skeleton",
    13: "Make Stone Pickaxe",
    14: "Make Stone Sword",
    15: "Wake Up",
    16: "Place Furnace",
    17: "Collect Coal",
    18: "Collect Iron",
    19: "Collect Diamond",
    20: "Make Iron Pickaxe",
    21: "Make Iron Sword",
    22: "Make Arrow",
    23: "Make Torch",
    24: "Place Torch",
    25: "Make Diamond Sword",
    26: "Make Iron Armour",
    27: "Make Diamond Armour",
    28: "Enter Gnomish Mines",
    29: "Enter Dungeon",
    30: "Enter Sewers",
    31: "Enter Vault",
    32: "Enter Troll Mines",
    33: "Enter Fire Realm",
    34: "Enter Ice Realm",
    35: "Enter Graveyard",
    36: "Defeat Gnome Warrior",
    37: "Defeat Gnome Archer",
    38: "Defeat Orc Solider",
    39: "Defeat Orc Mage",
    40: "Defeat Lizard",
    41: "Defeat Kobold",
    42: "Defeat Troll",
    43: "Defeat Deep Thing",
    44: "Defeat Pigman",
    45: "Defeat Fire Elemental",
    46: "Defeat Frost Troll",
    47: "Defeat Ice Elemental",
    48: "Damage Necromancer",
    49: "Defeat Necromancer",
    50: "Eat Bat",
    51: "Eat Snail",
    52: "Find Bow",
    53: "Fire Bow",
    54: "Collect Sapphire",
    55: "Learn Fireball",
    56: "Cast Fireball",
    57: "Learn Iceball",
    58: "Cast Iceball",
    59: "Collect Ruby",
    60: "Make Diamond Pickaxe",
    61: "Open Chest",
    62: "Drink Potion",
    63: "Enchant Sword",
    64: "Enchant Armour",
    65: "Defeat Knight",
    66: "Defeat Archer",
}

CRAFTAX_FULL_BACKDROP_BLOCK_TYPES = [
    "grass",
    "sand",
    "path",
    "fire_grass",
    "ice_grass",
    "gravel",
]
CRAFTAX_FULL_LOW_SALIENCE_OBJECTS = [
    "stone",
    "tree",
    "wood",
    "plant",
    "wall",
    "darkness",
    "wall_moss",
    "stalagmite",
    "fire_tree",
    "ice_shrub",
    "grave",
    "grave2",
    "grave3",
]
CRAFTAX_FULL_LOW_SALIENCE_MOBS = []
CRAFTAX_FULL_LOW_SALIENCE_ITEMS = []
CRAFTAX_FULL_HIGH_SALIENCE_OBJECTS = [
    "water",
    "lava",
    "coal",
    "iron",
    "diamond",
    "sapphire",
    "ruby",
    "crafting_table",
    "furnace",
    "ripe_plant",
    "chest",
    "fountain",
    "enchantment_table_fire",
    "enchantment_table_ice",
    "necromancer",
    "necromancer_vulnerable",
]
CRAFTAX_FULL_HIGH_SALIENCE_ITEMS = ["ladder_down", "ladder_up"]
CRAFTAX_FULL_HIGH_SALIENCE_MOBS = [
    "Zombie",
    "Gnome Warrior",
    "Orc Soldier",
    "Lizard",
    "Knight",
    "Troll",
    "Pigman",
    "Frost Troll",
    "Cow",
    "Bat",
    "Snail",
    "Skeleton",
    "Gnome Archer",
    "Orc Mage",
    "Kobold",
    "Archer",
    "Deep Thing",
    "Fire Elemental",
    "Ice Elemental",
    "Arrow",
    "Dagger",
    "Fireball",
    "Iceball",
    "Slimeball",
    "Arrow (Player)",
    "Dagger (Player)",
    "Fireball (Player)",
    "Iceball (Player)",
    "Slimeball (Player)",
]

FULL_TUTORIAL = """
# Craftax Wiki: Tutorial

This page explains how to beat the game!

## Basic Mechanics
Craftax is a game about exploring dungeons, mining, crafting and fighting enemies.
The player can move in the four cardinal directions using WASD and can interact using SPACE.
Interacting can cause the player to attempt to mine (a block), attack (a creature), drink (water or from a fountain), eat (fruit) or open a chest.

The player has 5 'intrinsics': health, hunger, thirst, energy and mana (magical energy).
Hunger, thirst and energy will naturally decrease and must be replenished by eating, drinking and sleeping respectively.
Mana is used for casting spells or enchanting items and will naturally recover.
Health will recover when hunger, thirst and energy are non-zero and will decrease if any of these are 0.
If the players health falls beneath 0 they will die and the game will restart.

To progress through the game the player needs to find the ladder on each floor, which can be used to descend to the next level.
Each floor possesses unique challenges and creatures, increasing in difficulty until the final boss level.
The ladders begin closed and the player must kill 8 creatures on each level to open up the respective ladders (with the exception of the overworld).
There are 9 levels in total.

## Floor 0: Overworld

The overworld consists of grasslands, lakes and mountains.
The player should begin by mining trees, before making a crafting table and using it to craft wooden tools by pressing the appropriate crafting keys while adjacent to the placed table.
You can then use the wooden pickaxe to mine stone and craft stone tools.
With stone tools, the player can mine coal and iron.
By placing a furnace next to the crafting table and standing adjacent to both (diagonal counts) the player can craft iron tools using wood, iron and coal.
Keep up your food by eating cows, make sure to drink water from the lakes and sleep when tired.
Bear in mind that the player is very vulnerable when sleeping so you should block yourself in by placing stone around you to sleep.

Now the player should find the first ladder down to the dungeon.  Make sure to collect lots of wood from the overworld as it is rarer in the dungeons.
The player might also want to collect seeds by pressing space on grass and plant these.  If the player returns to the overworld they will then have some easy food.

## Floor 1: Dungeon

Stand on the downward ladder and press the DESCEND key to climb down to the dungeon.
Note that you will appear on an upwards ladder, which can be climbed with the ASCEND key to return to the overworld.

The dungeon consists of a set of rooms joined by paths.  Rooms can contain fountains to drink from and chests with random items.
The first chest opened will contain a bow.  Arrows can be crafted at a crafting table using wood and stone.
The dungeon is inhabited by orc warriors and mages.  Once 8 of these have been killed the ladder to the next floor will open up.
The floor also contains snails which can be eaten.

If the player finds themselves on low health, they should block themselves in and either sleep or (if already at maximum energy) rest.
Resting causes the player to execute no-op actions until an intrinsic decays to 0, the player is attacked or the player recovers to full health.

### Attributes
Upon descending to each floor for the first time the player will be awarded an experience point.
These can be assigned to your attributes by pressing the appropriate key.
Each attribute starts at level 1 and can be upgraded to a maximum of level 5, with assignment being permanent.

**Dexterity**: Dexterity increases your maximum food, water and energy reserves, as well as making them decay slower.  It also increase damage done with a bow.

**Strength**: Strength increases your melee damage and maximum health.

**Intelligence**:  Intelligence increases your maximum mana, reduces mana decay, increases damage from spells and increases effectiveness of enchantments.

### Potions
The player will find potions of six different colours in the chests.
Potions will provide either +8 or -3 of either health, mana or energy.
However, the effects of each colour of potions is permuted every time the game is run, so the player will need to perform trial and error to figure out the ordering each game.

## Floor 2: Gnomish Mines

The gnomish mines are the first dark level.  To see on this level the player will need to place torches, which can be crafted with wood and coal.
There are pools of water to drink from and bats to eat.
The edges of the caverns are rich in ores to mine and should be taken advantage of.
The gnomes are stronger than the orcs and care must be taken to avoid being surrounded in the open spaces.
As well as coal and iron the player might find diamonds, sapphires and rubies.
Diamonds can be used to craft a sword (requires 2) or a pickaxe (requires 3) which can be used to mine sapphires and rubies.
The player can also craft iron armour (3 iron and 3 coal per piece) or diamond armour (3 diamonds) to reduce damage.

## Floor 3: Sewers

The sewers are similar to the dungeons in layout.  There are patches of water which need to be filled by placing stone and then mining it.
The lizards that inhabit this level are very dangerous and can swim through the water, while the Kobolds through high-damage daggers.

### Spellcasting
The first opened chest on this level will include a book, which can be read to learn a random spell (either fireball or iceball).
These are ranged attacks that consume 2 mana to cast.
Until now the only possible damage type has been 'physical', however the spells do fire and ice damage respectively.
Creatures on later levels have high resistance to physical damage and require fire or ice damage to kill.

### Enchanting
The sewers will contain the ice enchantment table.  The player can enchant their sword, armour or bow by standing next to the enchantment table and consuming 9 mana as well as the appropriate gemstone.
Sapphires are used for ice enchantments and rubies for fire enchantments.
An enchantment on the sword or bow will cause the player to deal +50% damage of the respective type.
An enchantment on armour will reduce damage of that type by 20% for each armour piece.

## Floor 4: Vaults
The vaults are another dungeon level.  Another book will be found along with the fire enchantment table.
The knights and archers on this floor are armoured and physical damage is halved, so using enchantments or spells is recommended.

## Floor 5: Troll Mines
The troll mines are a dark cavern-like level similar to the Gnomish Mines.
This level is the richest with ores so the player should make use of these and ideally craft full diamond armour.
The trolls are strong enemies and do a lot of damage.
The deep things in the water are very weak if you can hit them but do a lot of damage with their ranged attacks.

## Floor 6: Fire Realm
The fire realm consists of a set of islands separated by lava.
There is no way to obtain water on this level so the player may have to periodically return to the troll mines to drink.
The pig men and fire elementals are entirely resistant to fire damage and resistant to almost all physical damage, so ice damage (either through spells or enchantments) will probably be required to kill them.
Note that the player can build bridges across the lava by placing and mining stone.
There are also lots of coal and rubies to be found on this level.

## Floor 7: Ice Realm
The ice realm is a dark level inhabited by frost trolls and ice elementals.
These are the strongest creatures in the game and require fire damage to kill them.
There is no food on this level so the player may have to return to the fire realm to eat.
There are also lots of sapphires and rubies in the mountains.

## Floor 8: Graveyard
The graveyard is home to the necromancer who is the final enemy in the game.
There is no ladder out of the boss level.
There is no food or water on the level but the players hunger, thirst and energy will not decay on this level.
The necromancer will summon waves of enemies from the graves.
Once the player has defeated a wave of enemies the necromancer will enter his 'vulnerable' state, at which point the player can attack him.
Doing so will trigger the next wave of enemies.
Each wave of enemies corresponds to the creatures from a particular floor of the game.
Once the final wave (the ice realm wave) has been defeated the player can attack the necromancer and win the game!"""



import math
from typing import Dict, List, Literal

import jax
import jax.numpy as jnp
import numpy as np
from craftax.craftax.constants import *
from craftax.craftax.craftax_state import EnvState
from craftax.craftax_classic.renderer import render_craftax_pixels
from craftax.craftax_env import make_craftax_env_from_name

from craftaxlm.classic.metadata import (
    CRAFTAX_CLASSIC_ACHIEVEMENTS,
    CRAFTAX_CLASSIC_ACTION_MAPPING,
)
from craftaxlm.classic.state import (
    CraftaxClassicState,
    CraftaxRecorder,
    mob_id_to_name,
    player_chars,
    render_craftax_classic_text_custom,
)
from craftaxlm.shared import CraftaxBaseACI


class CraftaxClassicACI(CraftaxBaseACI):
    formatting: Literal["md", "xml"] = "md"
    map_format: Literal["full", "compact"] = "full"
    vantage: int = 3

    def __init__(
        self, formatting="md", map_format="full", vantage=3, verbose=False, seed=None
    ):
        self.formatting = formatting
        self.map_format = map_format
        self.vantage = vantage
        self.verbose = verbose
        self.achievement_deltas = []
        super().__init__(seed=seed)

    def make_env(self):
        return make_craftax_env_from_name(
            "Craftax-Classic-Symbolic-v1", auto_reset=False
        )

    def create_starting_obs(self):
        step_info = self.create_step_info(self.state, 0.0, False)
        return {
            "state_image": step_info["state_image"],
            "state_text": step_info["state_text"],
            "reward": 0.0,
            "done": False,
        }

    def map_action_string_to_int(self, action_string: str) -> int:
        return CRAFTAX_CLASSIC_ACTION_MAPPING.get(action_string.lower(), 0)

    def get_achievements(self, state):
        return {
            "achievements": {
                k: state.achievements[i]
                for i, k in CRAFTAX_CLASSIC_ACHIEVEMENTS.items()
            }
        }

    def get_achievement_delta(self, achievements):
        """
        Calculate which achievements have been newly unlocked.

        Args:
            achievements: Dictionary of current achievements

        Returns:
            Dictionary of newly unlocked achievements or None if no new achievements
        """
        if not hasattr(self, "last_achievements"):
            self.last_achievements = {
                "achievements": {k: 0 for k in achievements["achievements"].keys()}
            }
            return {}

        # Find new achievements (changed from 0 to 1)
        delta = {}
        for k, v in achievements["achievements"].items():
            if v > 0 and self.last_achievements["achievements"].get(k, 0) == 0:
                delta[k] = str(v)

        # Update last_achievements
        self.last_achievements = achievements

        # Return None if no new achievements
        if not delta:
            return {}

        return {"new_achievements": delta}

    def create_step_info(self, state, reward, done):
        craftaxlm_state = render_craftax_classic_text_custom(
            state, vantage=self.vantage
        )
        return {
            "state_image": craftaxlm_state.image,
            "state_text": craftaxlm_state.render_to_text_simple(
                verbose=self.verbose,
                formatting=self.formatting,
                map_format=self.map_format,
            ),
            "reward": float(reward),
            "done": bool(done),
        }

    def accept_tool_call(self, tool_call):
        """
        Accepts an OpenAI SDK-like tool call and executes the corresponding action.

        Args:
            tool_call: A dictionary with the structure similar to OpenAI tool calls:
                {
                    "type": "function",
                    "function": {
                        "name": "take_action",
                        "arguments": {
                            "action": "up" | "down" | "left" | "right" | etc.
                        }
                    },
                    "id": "call_abc123" # optional tool call ID
                }

        Returns:
            Dictionary formatted as an OpenAI tool message response:
            {
                "role": "tool",
                "content": JSON string with state information,
                "tool_call_id": The ID of the tool call being responded to
            }
        """
        state_before = self.state

        # Extract action from tool call
        action = None
        tool_call_id = None

        if isinstance(tool_call, str):
            # Handle legacy string action format
            action = tool_call
        elif isinstance(tool_call, dict):
            # Extract action from OpenAI SDK-like tool call format
            if "id" in tool_call:
                tool_call_id = tool_call["id"]

            if "function" in tool_call:
                function_data = tool_call["function"]
                if "arguments" in function_data:
                    arguments = function_data["arguments"]
                    # Handle both string and dict arguments
                    if isinstance(arguments, str):
                        import json

                        try:
                            arguments = json.loads(arguments)
                        except json.JSONDecodeError:
                            action = arguments

                    if isinstance(arguments, dict) and "action" in arguments:
                        action = arguments["action"]

        if action is None:
            raise ValueError(f"Could not extract action from tool call: {tool_call}")

        # Map action string to integer and execute
        action_int = self.map_action_string_to_int(action)
        _, state, reward, done, info = self.env.step(
            self.rngs[2], state_before, action_int, self.env_params
        )
        self.state = state

        # Calculate achievement deltas
        achievements = self.get_achievements(state)
        achievement_delta = self.get_achievement_delta(achievements)
        assert achievement_delta is not None
        # Store achievement deltas if the attribute exists
        # if hasattr(self, "achievement_deltas"):
        self.achievement_deltas.append(achievement_delta)

        # Print if verbose mode is enabled
        # if achievement_delta and hasattr(self, "verbose") and self.verbose:
        print(achievement_delta)

        step_info = self.create_step_info(state, reward, done)
        updates, expectation = get_action_result_vs_expectation(
            state_before, state, action
        )

        # Format response as an OpenAI tool message
        import base64
        import io
        import json

        import numpy as np
        from PIL import Image

        # Convert numpy array to base64 for OpenAI compatibility
        state_image = step_info.pop("state_image")  # Remove image from step_info

        # Convert numpy array to PIL Image and then to bytes
        pil_image = Image.fromarray(state_image)
        buffer = io.BytesIO()
        pil_image.save(buffer, format="PNG")

        # Get base64 encoded string
        img_str = base64.b64encode(buffer.getvalue()).decode("utf-8")

        # Format as OpenAI compatible image object
        image_obj = {
            "type": "image_url",
            "image_url": {"url": f"data:image/png;base64,{img_str}"},
        }

        response_content = {
            "direct_results": updates,
            "expected_results": expectation,
            "state": step_info,
            "state_image": image_obj,
        }

        # Include achievement delta if available
        if achievement_delta:
            response_content["achievement_delta"] = achievement_delta

        response = {
            "role": "tool",
            "content": json.dumps(response_content),
        }

        # Add tool_call_id if it was provided
        if tool_call_id:
            response["tool_call_id"] = tool_call_id

        return response

    def step_action(self, action_string, tool_call_id=None):
        """
        Helper method to perform an action using the OpenAI tool call format.

        Args:
            action_string: String representation of the action to take
            tool_call_id: Optional ID for the tool call

        Returns:
            Dictionary formatted as an OpenAI tool message response
        """
        # Create a tool call in OpenAI format
        tool_call = {
            "type": "function",
            "function": {"name": "take_action", "arguments": {"action": action_string}},
        }

        # Add tool_call_id if provided
        if tool_call_id:
            tool_call["id"] = tool_call_id

        # Process the tool call
        return self.accept_tool_call(tool_call)


def render_craftax_text(state):
    """Renders the CraftAX Classic state as text representation."""
    obs_dim_array = jnp.array([OBS_DIM[0], OBS_DIM[1]], dtype=jnp.int32)

    # Map
    padded_grid = jnp.pad(
        state.map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=BlockType.OUT_OF_BOUNDS.value,
    )

    tl_corner = state.player_position - obs_dim_array // 2 + MAX_OBS_DIM + 2
    map_view = jax.lax.dynamic_slice(padded_grid, tl_corner, OBS_DIM)

    # Convert blocks to text representation
    block_chars = {
        BlockType.INVALID.value: ".",
        BlockType.OUT_OF_BOUNDS.value: "#",
        BlockType.GRASS.value: "G",
        BlockType.WATER.value: "~",
        BlockType.STONE.value: "S",
        BlockType.TREE.value: "T",
        BlockType.WOOD.value: "W",
        BlockType.PATH.value: "P",
        BlockType.COAL.value: "C",
        BlockType.IRON.value: "I",
        BlockType.DIAMOND.value: "D",
        BlockType.CRAFTING_TABLE.value: "B",  # B for crafting Bench
        BlockType.FURNACE.value: "F",
        BlockType.SAND.value: "_",
        BlockType.LAVA.value: "L",
        BlockType.PLANT.value: "p",
        BlockType.RIPE_PLANT.value: "r",
        BlockType.WALL.value: "X",
        BlockType.DARKNESS.value: " ",
        BlockType.WALL_MOSS.value: "M",
        BlockType.STALAGMITE.value: "^",
        BlockType.SAPPHIRE.value: "$",
        BlockType.RUBY.value: "R",
        BlockType.CHEST.value: "H",
        BlockType.FOUNTAIN.value: "O",
        BlockType.FIRE_GRASS.value: "f",
        BlockType.ICE_GRASS.value: "i",
        BlockType.GRAVEL.value: "g",
        BlockType.FIRE_TREE.value: "t",
        BlockType.ICE_SHRUB.value: "s",
        BlockType.ENCHANTMENT_TABLE_FIRE.value: "E",
        BlockType.ENCHANTMENT_TABLE_ICE.value: "e",
        BlockType.NECROMANCER.value: "N",
        BlockType.GRAVE.value: "v",
        BlockType.GRAVE2.value: "V",
        BlockType.GRAVE3.value: "u",
        BlockType.NECROMANCER_VULNERABLE.value: "n",
    }

    # Create text grid - swap x and y to match game coordinates
    height, width = map_view.shape
    text_grid = []
    for y in range(height):
        row = []
        for x in range(width):
            block_value = int(map_view[y, x])
            row.append(block_chars.get(block_value, "?"))
        text_grid.append(row)

    # Add player at center
    center_y, center_x = obs_dim_array // 2
    player_direction = int(state.player_direction)
    # print("Player direction: ", player_direction)
    text_grid[center_y][center_x] = player_chars.get(player_direction, "P")

    # print("Player char: ", text_grid[center_y][center_x])
    # Add mobs - ensure coordinates are properly aligned
    def add_mob_to_grid(mob_positions, mob_masks, symbol):
        for pos, mask in zip(mob_positions, mob_masks):
            if not mask:
                continue
            # Calculate local position relative to player
            local_pos = pos - state.player_position + obs_dim_array // 2
            # Check if mob is within visible area
            if (local_pos >= 0).all() and (local_pos < obs_dim_array).all():
                y, x = int(local_pos[0]), int(local_pos[1])
                text_grid[y][x] = symbol

    add_mob_to_grid(state.zombies.position, state.zombies.mask, "Z")
    add_mob_to_grid(state.cows.position, state.cows.mask, "c")
    add_mob_to_grid(state.skeletons.position, state.skeletons.mask, "K")
    add_mob_to_grid(state.arrows.position, state.arrows.mask, "a")

    # Convert to string - join rows in correct order
    text_map = "\n".join("".join(row) for row in text_grid)

    return text_map


compact_map_translation_full = {
    # Basic blocks
    "G": "Grass",
    "~": "Water",
    "S": "Stone",
    "T": "Tree",
    "W": "Wood",
    "P": "Path",
    "C": "Coal",
    "I": "Iron",
    "D": "Diamond",
    "B": "Crafting Bench",
    "F": "Furnace",
    "E": "Enchantment Table Fire",
    "e": "Enchantment Table Ice",
    # Additional blocks
    "_": "Sand",
    "L": "Lava",
    "p": "Plant",
    "r": "Ripe Plant",
    "X": "Wall",
    " ": "Darkness",
    "M": "Wall Moss",
    "^": "Stalagmite",
    "$": "Sapphire",
    "R": "Ruby",
    "H": "Chest",
    "O": "Fountain",
    # Special terrain
    "f": "Fire Grass",
    "i": "Ice Grass",
    "g": "Gravel",
    "t": "Fire Tree",
    "s": "Ice Shrub",
    # Necromancer and graves
    "N": "Necromancer",
    "n": "Necromancer Vulnerable",
    "v": "Grave",
    "V": "Grave2",
    "u": "Grave3",
    # Mobs
    "Z": "Zombie",
    "c": "Cow",
    "K": "Skeleton",
    "a": "Arrow",
    # Boundaries
    ".": "Invalid",
    "#": "Out of Bounds",
}

compact_map_translation_classic = {
    # Basic blocks
    "G": "Grass",
    "~": "Water",
    "S": "Stone",
    "T": "Tree",
    "W": "Wood",
    "P": "Path",
    "C": "Coal",
    "I": "Iron",
    "D": "Diamond",
    "B": "Crafting Bench",
    "F": "Furnace",
    "E": "Enchantment Table Fire",
    "e": "Enchantment Table Ice",
    # Additional blocks
    "_": "Sand",
    "L": "Lava",
    "p": "Plant",
    "r": "Ripe Plant",
    "X": "Wall",
    # Mobs
    "Z": "Zombie",
    "c": "Cow",
    "K": "Skeleton",
    "a": "Arrow",
    # Boundaries
    ".": "Invalid",
    "#": "Out of Bounds",
}


def render_craftax_classic_text_exp(
    state: EnvState, enable_recording=False, vantage=3
) -> CraftaxClassicState:
    print("Exp")
    map_data = []
    for x in range(state.map.shape[0]):
        for y in range(state.map.shape[1]):
            if (
                not max(
                    abs(x - state.player_position[0]), abs(y - state.player_position[1])
                )
                <= 4
            ):
                continue
            tile = {
                "position": {
                    "x": x - state.player_position[0],
                    "y": y - state.player_position[1],
                },
                "visible": max(
                    abs(x - state.player_position[0]), abs(y - state.player_position[1])
                )
                <= 4,
                "block": BlockType(state.map[x, y]).name.lower(),
            }
            if state.mob_map[x, y].max() > 0.5:
                tile["mob"] = mob_id_to_name(state.mob_map[x, y].argmax())
            map_data.append(tile)
    text_map = render_craftax_text(state)

    inventory_data = {
        "resources": {
            "wood": int(state.inventory.wood),
            "stone": int(state.inventory.stone),
            "coal": int(state.inventory.coal),
            "iron": int(state.inventory.iron),
            "diamond": int(state.inventory.diamond),
            "sapling": int(state.inventory.sapling),
        },
        "tools": {},
    }

    inventory_data["tools"]["pickaxe"] = {
        "wood": int(state.inventory.wood_pickaxe),
        "stone": int(state.inventory.stone_pickaxe),
        "iron": int(state.inventory.iron_pickaxe),
    }

    inventory_data["tools"]["sword"] = {
        "wood": int(state.inventory.wood_sword),
        "stone": int(state.inventory.stone_sword),
        "iron": int(state.inventory.iron_sword),
    }

    player_data = {
        "health": int(state.player_health),
        "food": int(state.player_food),
        "drink": int(state.player_drink),
        "energy": int(state.player_energy),
        "direction_facing": Action(state.player_direction).name.lower(),
    }

    environment_data = {
        "light_level": float(state.light_level),
        "is_sleeping": bool(state.is_sleeping),
        "floor": 0,  # Hardcode to 0 since player_level isn't in EnvState
    }

    def to_json_friendly(data):
        if isinstance(data, dict):
            return {key: to_json_friendly(value) for key, value in data.items()}
        elif isinstance(data, (list, tuple)):
            return [to_json_friendly(item) for item in data]
        elif isinstance(data, (jnp.ndarray, np.ndarray)):
            if data.ndim == 0:
                return data.item()
            return [to_json_friendly(item) for item in data]
        elif isinstance(data, (jnp.bool_, np.bool_)):
            return bool(data)
        elif isinstance(data, (jnp.integer, np.integer)):
            return int(data)
        elif isinstance(data, (jnp.floating, np.floating)):
            return float(data)
        else:
            return data

    pixels = render_craftax_pixels(state, BLOCK_PIXEL_SIZE_HUMAN)
    # Convert from JAX array to numpy and ensure uint8 format
    frame = np.array(pixels).astype(np.uint8)

    craftax_state = CraftaxClassicState(
        map_full=to_json_friendly(map_data),
        map_compact=text_map,
        inventory=to_json_friendly(inventory_data),
        player=to_json_friendly(player_data),
        environment=to_json_friendly(environment_data),
        image=frame,
        recorder=None if not enable_recording else CraftaxRecorder(),
        vantage=vantage,
    )

    # Record frame if recording is enabled
    if enable_recording:
        craftax_state.recorder.record_frame(state)

    return craftax_state


def canonical_crafter_score_classic(achievements_by_run: List[Dict]):
    assert all(
        isinstance(achievements, dict)
        and all(
            k in achievements
            for k in [
                "Collect Wood",
                "Place Table",
                "Eat Cow",
                "Collect Sapling",
                "Collect Drink",
                "Make Wood Pickaxe",
                "Make Wood Sword",
                "Place Plant",
                "Defeat Zombie",
                "Collect Stone",
                "Place Stone",
                "Eat Plant",
                "Defeat Skeleton",
                "Make Stone Pickaxe",
                "Make Stone Sword",
                "Wake Up",
                "Place Furnace",
                "Collect Coal",
                "Collect Iron",
                "Collect Diamond",
                "Make Iron Pickaxe",
                "Make Iron Sword",
            ]
        )
        and all(isinstance(v, int) for v in achievements.values())
        for achievements in achievements_by_run
    )
    successes_by_achievement = {
        achievement: sum(1 for run in achievements_by_run if run[achievement] > 0)
        for achievement in achievements_by_run[0].keys()
    }
    success_rates = [
        count / len(achievements_by_run) for count in successes_by_achievement.values()
    ]

    # Compute the Crafter score: S = exp((1/N)*sum(ln(1 + s_i))) - 1, where s_i is the success rate for achievement i.
    N = len(success_rates)
    return math.exp(sum(math.log(1 + s) for s in success_rates) / N) - 1


CRAFTAX_CLASSIC_ACTION_MAPPING = {
    "noop": 0,
    "left": 1,
    "right": 2,
    "up": 3,
    "down": 4,
    "do": 5,
    "sleep": 6,
    "place_stone": 7,
    "place_table": 8,
    "place_furnace": 9,
    "place_plant": 10,
    "make_wood_pickaxe": 11,
    "make_stone_pickaxe": 12,
    "make_iron_pickaxe": 13,
    "make_wood_sword": 14,
    "make_stone_sword": 15,
    "make_iron_sword": 16,
}


def get_action_expectations():
    """
    Returns a dictionary containing the required and optional expected results
    for each valid action in CraftAx Classic, using parametric inventory format.
    """
    expectations = {
        "noop": {"required": [], "optional": []},
        "left": {"required": [], "optional": ["turn_left"]},
        "right": {"required": [], "optional": ["turn_right"]},
        "up": {"required": [], "optional": ["turn_up"]},
        "down": {"required": [], "optional": ["turn_down"]},
        "do": {
            "required": [],
            "optional": [
                "stone_k->[k+1]",
                "wood_k->[k+1]",
                "coal_k->[k+1]",
                "iron_k->[k+1]",
                "diamond_k->[k+1]",
                "stone_removed",
                "tree_removed",
                "coal_removed",
                "iron_removed",
                "diamond_removed",
                "zombie_removed",
                "skeleton_removed",
            ],
        },
        "sleep": {"required": [], "optional": ["time_night->day", "health_low->full"]},
        "place_stone": {"required": ["stone_k->[k-1]", "stone_added"], "optional": []},
        "place_table": {
            "required": ["wood_k->[k-2]", "crafting_table_added"],
            "optional": [],
        },
        "place_furnace": {
            "required": ["stone_k->[k-1]", "furnace_added"],
            "optional": [],
        },
        "place_plant": {"required": [], "optional": []},
        "make_wood_pickaxe": {
            "required": ["wood_k->[k-1]", "wood_pickaxe_k->[k+1]"],
            "optional": [],
        },  # update these?
        "make_stone_pickaxe": {
            "required": ["stone_k->[k-1]", "wood_k->[k-1]", "stone_pickaxe_k->[k+1]"],
            "optional": [],
        },
        "make_iron_pickaxe": {
            "required": ["iron_k->[k-1]", "wood_k->[k-1]", "iron_pickaxe_k->[k+1]"],
            "optional": [],
        },
        "make_wood_sword": {
            "required": ["wood_k->[k-1]", "wood_sword_k->[k+1]"],
            "optional": [],
        },
        "make_stone_sword": {
            "required": ["stone_k->[k-1]", "wood_k->[k-1]", "stone_sword_k->[k+1]"],
            "optional": [],
        },
        "make_iron_sword": {
            "required": ["iron_k->[k-1]", "wood_k->[k-1]", "iron_sword_k->[k+1]"],
            "optional": [],
        },
    }

    return expectations


def get_action_expectation(action):
    """
    Returns the expected results for a specific action.

    Args:
        action: String name or integer ID of the action

    Returns:
        Dictionary with 'required' and 'optional' effects
    """
    expectations = get_action_expectations()

    # Handle case where action is passed as an integer ID
    if isinstance(action, int):
        for name, id in CRAFTAX_CLASSIC_ACTION_MAPPING.items():
            if id == action:
                action = name
                break

    if action in expectations:
        return expectations[action]
    else:
        return {"required": [], "optional": []}


def get_action_result(state_before, state_after):
    # updates = {}

    # Position
    updates = []
    player_position_before = state_before.player_position
    player_position_after = state_after.player_position
    # print("Player position before: ", player_position_before)
    # print("Player position after: ", player_position_after)
    if not all(player_position_before[i] == player_position_after[i] for i in range(2)):
        if player_position_before[0] < player_position_after[0]:
            updates.append("move_down")
        elif player_position_before[0] > player_position_after[0]:
            updates.append("move_up")
        elif player_position_before[1] < player_position_after[1]:
            updates.append("move_right")
        elif player_position_before[1] > player_position_after[1]:
            updates.append("move_left")
    else:
        pass
        # print("Player position is the same")

    # Direction
    direction_dict = {
        3: "up",
        4: "down",
        1: "left",
        2: "right",
    }
    player_direction_before = state_before.player_direction
    player_direction_after = state_after.player_direction
    if player_direction_before != player_direction_after:
        print(player_direction_after)
        updates.append("turn_" + direction_dict[int(player_direction_after)])

    # Inventory
    craftax_state_before = render_craftax_classic_text_custom(state_before, vantage=3)
    craftax_state_after = render_craftax_classic_text_custom(state_after, vantage=3)
    # print("Keys: ",craftax_state_before.inventory)
    # for rt in craftax_state_before.inventory.keys():
    for k in craftax_state_before.inventory["resources"].keys():
        if (
            craftax_state_before.inventory["resources"][k]
            != craftax_state_after.inventory["resources"][k]
        ):
            updates.append(
                k
                + "_"
                + str(craftax_state_before.inventory["resources"][k])
                + "->"
                + str(craftax_state_after.inventory["resources"][k])
            )
    for tool_type in craftax_state_before.inventory["tools"].keys():
        for tool_variety in craftax_state_before.inventory["tools"][tool_type].keys():
            if (
                craftax_state_before.inventory["tools"][tool_type][tool_variety]
                != craftax_state_after.inventory["tools"][tool_type][tool_variety]
            ):
                updates.append(
                    tool_variety
                    + "_"
                    + tool_type
                    + "_"
                    + str(
                        craftax_state_before.inventory["tools"][tool_type][tool_variety]
                    )
                    + "->"
                    + str(
                        craftax_state_after.inventory["tools"][tool_type][tool_variety]
                    )
                )
    # Map
    # If these are removed, show it: stone, tree, coal, iron, diamond, zombie, skeleton
    # If these are added, show it: crafting_table, furnace
    if all(player_position_before[i] == player_position_after[i] for i in range(2)):
        for tile in craftax_state_before.map_full:
            after_tile = next(
                (
                    t
                    for t in craftax_state_after.map_full
                    if t["position"]["x"] == tile["position"]["x"]
                    and t["position"]["y"] == tile["position"]["y"]
                ),
                None,
            )
            if tile["block"] in [
                "stone",
                "tree",
                "coal",
                "iron",
                "diamond",
                "zombie",
                "skeleton",
            ]:
                if tile != after_tile:
                    updates.append(tile["block"] + "_removed")
        for tile in craftax_state_after.map_full:
            before_tile = next(
                (
                    t
                    for t in craftax_state_before.map_full
                    if t["position"]["x"] == tile["position"]["x"]
                    and t["position"]["y"] == tile["position"]["y"]
                ),
                None,
            )
            if tile["block"] in ["crafting_table", "furnace"]:
                if tile != before_tile:
                    updates.append(tile["block"] + "_added")
    return updates


def get_action_result_vs_expectation(state_before, state_after, action):
    updates = get_action_result(state_before, state_after)
    expectation = get_action_expectation(action)
    return updates, expectation


def ensure_json_serializable(obj):
    """Convert NumPy arrays and other non-serializable objects to JSON-compatible types."""
    import numpy as np

    if isinstance(obj, dict):
        return {k: ensure_json_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [ensure_json_serializable(item) for item in obj]
    elif isinstance(obj, np.ndarray):
        # Convert NumPy array to Python native type
        return obj.item() if obj.size == 1 else obj.tolist()
    elif isinstance(obj, np.bool_):
        return bool(obj)
    elif hasattr(obj, "dtype") and np.isscalar(obj):
        # Handle other NumPy scalar types
        return obj.item()
    else:
        return obj


if __name__ == "__main__":
    import json

    craftax_aci = CraftaxClassicACI(
        formatting="md", map_format="full", verbose=True, seed=0
    )
    craftax_aci.reset()

    # Example 1: Using direct step calls (old way)
    print("EXAMPLE 1: Using direct step calls")
    state_before = craftax_aci.state
    action_int = craftax_aci.map_action_string_to_int("down")
    print("Action int: ", action_int)
    _, state, reward, done, info = craftax_aci.env.step(
        craftax_aci.rngs[2], state_before, action_int, craftax_aci.env_params
    )
    updates, expected_updates = get_action_result_vs_expectation(
        state_before, state, action_int
    )
    craftax_aci.state = state  # Make sure to update the state
    print("Updates:", updates)
    print("Expected:", expected_updates)

    # Example 2: Using the new step_action method
    print("\nEXAMPLE 2: Using step_action method")
    response = craftax_aci.step_action("down", tool_call_id="action_123")
    print("Tool Role:", response["role"])
    print("Tool Call ID:", response.get("tool_call_id"))

    content = json.loads(response["content"])
    print("Achievement Delta:", content.get("achievement_delta"))

    # Let's try to trigger an achievement
    print("\nTrying to trigger an achievement...")

    # Find a tree and try to collect wood (multiple steps to demonstrate)
    # First, look around for a tree
    directions = ["up", "right", "down", "left"]
    for direction in directions:
        print(f"\nChecking {direction} direction...")
        response = craftax_aci.step_action(direction)
        content = json.loads(response["content"])

        # See if we're facing a tree
        state_text = content["state"]["state_text"]
        if "Tree" in state_text and "Object_you_are_facing" in state_text:
            print(f"Found a tree! Attempting to collect wood...")
            # Use the 'do' action to collect wood
            response = craftax_aci.step_action("do")
            content = json.loads(response["content"])
            print("Action results:", content["direct_results"])

            # Check if we got an achievement
            if content.get("achievement_delta"):
                print("Achievement unlocked:", content["achievement_delta"])
            break

    # Example 3: Using accept_tool_call with OpenAI format
    print("\nEXAMPLE 3: Using accept_tool_call with OpenAI format")
    tool_call = {
        "type": "function",
        "function": {"name": "take_action", "arguments": {"action": "up"}},
        "id": "call_abc123",
    }
    response = craftax_aci.accept_tool_call(tool_call)
    print("Tool Role:", response["role"])
    print("Tool Call ID:", response.get("tool_call_id"))

    # Parse and display content for better readability
    import json

    content = json.loads(response["content"])
    print("\nResults:")
    print("Direct Results:", content["direct_results"])
    print("Expected Results:", content["expected_results"])
    print("Achievement Delta:", content.get("achievement_delta"))

    # Verify image format is correct
    image_data = content["state_image"]
    print("\nImage Data Format:")
    print(f"Type: {image_data['type']}")
    print(f"URL prefix: {image_data['image_url']['url'][:30]}...")

    # Optional: Save the image to a file to verify it works
    import base64

    try:
        # Extract base64 data
        base64_data = image_data["image_url"]["url"].split(",")[1]
        # Decode and save
        with open("test_output_image.png", "wb") as f:
            f.write(base64.b64decode(base64_data))
        print("\nImage saved to test_output_image.png")
    except Exception as e:
        print(f"Error saving image: {e}")

    # Show all accumulated achievement deltas
    print("\nAll Achievement Deltas:")
    print("Length of achievement deltas: ", len(craftax_aci.achievement_deltas))
    for i, delta in enumerate(craftax_aci.achievement_deltas):
        if delta:
            print(f"Step {i+1}:", delta)
    assert isinstance(craftax_aci.achievement_deltas, list)
    assert len(craftax_aci.achievement_deltas) > 0
    print("Achievement deltas: ", craftax_aci.achievement_deltas)



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

import imageio
import jax
import jax.numpy as jnp
import numpy as np
from craftax.craftax.constants import *
from craftax.craftax.constants import BLOCK_PIXEL_SIZE_HUMAN
from craftax.craftax.craftax_state import EnvState
from craftax.craftax_classic.renderer import render_craftax_pixels

from craftaxlm.classic.metadata import (
    CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES,
    CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS,
    CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS,
    CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS,
    CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS,
)
from craftaxlm.shared import CraftaxState, mob_id_to_name

player_chars = {1: "←", 2: "→", 3: "↑", 4: "↓"}


@dataclass
class CraftaxClassicState(CraftaxState):
    
    map_full: List[Dict]
    map_compact: str
    inventory: Dict
    player: Dict
    environment: Dict
    image: np.ndarray
    recorder: Optional["CraftaxRecorder"] = None
    vantage: int = 3

    @classmethod
    def create(cls, *args, enable_recording=False, **kwargs):
        recorder = CraftaxRecorder() if enable_recording else None
        return cls(*args, **kwargs, recorder=recorder)

    def record_if_enabled(self, state: EnvState):
        if self.recorder is not None:
            self.recorder.record_frame(state)

    def save_recording(self, filename="episode.mp4", fps=30):
        if self.recorder is not None:
            self.recorder.save_video(filename, fps)

    def render_map_to_text(self, ignore_distant_low_salience=True):
        unique_blocks = list(
            set([tile["block"] for tile in self.map_full if "block" in tile])
        )
        if not set(unique_blocks).issubset(
            set(
                CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES
                + CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS
                + CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS
            )
        ):
            raise ValueError(
                f"Unknown block types: {set(unique_blocks) - set(CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES + CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS + CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS)}"
            )
        unique_mobs = list(
            set([tile["mob"] for tile in self.map_full if "mob" in tile])
        )
        if not set(unique_mobs).issubset(
            set(CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS + CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS)
        ):
            raise ValueError(
                f"Unknown mob types: {set(unique_mobs) - set(CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS + CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS)}"
            )

        def count_nearby_blocks(center_x, center_y, radius):
            block_counts = {}
            for tile in self.map_full:
                dx = abs(tile["position"]["x"] - center_x)
                dy = abs(tile["position"]["y"] - center_y)
                if dx <= radius and dy <= radius:
                    if not "block" in tile:
                        continue
                    block_type = tile["block"]
                    if block_type in CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES:
                        block_counts[block_type] = block_counts.get(block_type, 0) + 1
            return block_counts

        center_x, center_y = 0, 0
        nearby_blocks = count_nearby_blocks(center_x, center_y, 2)
        backdrop = (
            max(nearby_blocks, key=nearby_blocks.get) if nearby_blocks else "unknown"
        )

        low_salience_objects = CRAFTAX_CLASSIC_LOW_SALIENCE_OBJECTS + [
            object
            for object in CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES
            if object != backdrop
        ]

        def describe_xy(x, y):
            # THIS IS SO CONFUSING - X is UP and DOWN, Y is LEFT and RIGHT
            description = ""
            if x > 0:
                description += f"{x} steps down"
            elif x < 0:
                description += f"{-x} steps up"
            if (x != 0) and (y != 0):
                description += " and "
            if y < 0:
                description += f"{-y} steps left"
            elif y > 0:
                description += f"{y} steps right"
            return description

        # periphery = []
        bound = self.vantage  # 6
        # Block
        # tile["block"].capitalize() + " is " + describe_xy(x, y)

        # Mob
        # periphery.append("A " + tile["mob"] + " is " + describe_xy(x, y))
        # for direction in [
        #     (-1, -1),
        #     (-1, 0),
        #     (-1, 1),
        #     (0, -1),
        #     (0, 1),
        #     (1, -1),
        #     (1, 0),
        #     (1, 1),
        # ]:
        #     found = False
        #     for distance in range(1, bound):
        #         if found:
        #             break
        #         x = center_x + direction[0] * distance
        #         y = center_y + direction[1] * distance
        #         for tile in self.map_full:
        #             if tile["position"]["x"] == x and tile["position"]["y"] == y:
        #                 if tile["block"] in low_salience_objects:
        #                     periphery.append(
        #                         tile["block"].capitalize() + " is " + describe_xy(x, y)
        #                     )
        #                     if ignore_distant_low_salience:
        #                         found = True
        #                 if (
        #                     "mob" in tile
        #                     and tile["mob"] in CRAFTAX_CLASSIC_LOW_SALIENCE_MOBS
        #                 ):
        #                     periphery.append(
        #                         "A " + tile["mob"] + " is " + describe_xy(x, y)
        #                     )
        #                     if ignore_distant_low_salience:
        #                         found = True

        # high_salience = []
        # for tile in self.map_full:
        #     if (
        #         tile["visible"]
        #         and tile["block"] in CRAFTAX_CLASSIC_HIGH_SALIENCE_OBJECTS
        #     ):
        #         high_salience.append(
        #             tile["block"].capitalize()
        #             + " is "
        #             + describe_xy(tile["position"]["x"], tile["position"]["y"]),
        #         )
        #     elif (
        #         tile["visible"]
        #         and "mob" in tile
        #         and tile["mob"] in CRAFTAX_CLASSIC_HIGH_SALIENCE_MOBS
        #     ):
        #         high_salience.append(
        #             "A "
        #             + tile["mob"]
        #             + " is "
        #             + describe_xy(tile["position"]["x"], tile["position"]["y"])
        #         )

        grid = []
        for x in range(-bound, bound + 1):
            row = []
            for y in range(-bound, bound + 1):
                row.append((center_x + x, center_y + y))
            grid.extend(row)
        surroundings = []
        def get_tile(x,y, map_full):
            # Find the tile with matching position coordinates
            for tile in map_full:
                if tile["position"]["x"] == x and tile["position"]["y"] == y:
                    # Return None if tile is not visible
                    if not tile["visible"]:
                        return None
                    return tile
            return None
        
        for coord in grid:
            x, y = coord
            if (x, y) == (0, 0):
                continue

            tile = get_tile(x, y, self.map_full)
            if tile is None:
                continue
            #print("Tile: ", tile)
            if "block" in tile and tile["block"] not in CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES:
                surroundings.append(
                    tile["block"].capitalize() + " is " + describe_xy(x, y)
                )
            if "mob" in tile:
                surroundings.append("A " + tile["mob"] + " is " + describe_xy(x, y))

        #print("Map full: ", self.map_full)

        direction = self.player["direction_facing"]
        direction_to_offset = {
            "up": (-1, 0),
            "down": (1, 0),
            "left": (0, -1),
            "right": (0, 1),
        }

        # Get the offset for the direction we're facing
        dx, dy = direction_to_offset.get(direction, (0, 0))

        # Look for objects exactly one step in the direction we're facing
        facing_objects = []
        for tile in self.map_full:
            if tile["position"]["x"] == dx and tile["position"]["y"] == dy:
                if (
                    "block" in tile
                    and tile["block"] not in CRAFTAX_CLASSIC_BACKDROP_BLOCK_TYPES
                ):
                    facing_objects.append(tile["block"].capitalize())
                if "mob" in tile:
                    facing_objects.append(f"A {tile['mob']}")

        # If no match found, we treat it as "No object"
        if not facing_objects:
            facing_objects = ["No object directly in front of you"]

        # print("Facing objects: ", facing_objects)
        return {
            "terrain_underneath_you": backdrop,
            "surroundings": surroundings,  # periphery + high_salience,
            "object_you_are_facing": facing_objects[0],
        }

    def render_inventory_to_text(self, include_absent_inventory=True):
        def process_inventory(inv_dict):
            processed = {}
            for key, value in inv_dict.items():
                if isinstance(value, dict):
                    processed_sub = process_inventory(value)
                    if processed_sub:
                        processed[key] = processed_sub
                elif include_absent_inventory or value > 0:
                    processed[key] = value
            return processed

        return process_inventory(self.inventory)

    def render_environment_to_text(self, include_absent_environment_attributes=True):
        attributes_to_hide = []
        environment = {}
        for key, value in self.environment.items():
            if key not in attributes_to_hide and (
                include_absent_environment_attributes or value
            ):
                environment[key] = value
        return environment


# def render_craftax_classic_text_custom(
#     state: EnvState, enable_recording=False
# ) -> CraftaxClassicState:
#     map_data = []
#     for x in range(state.map.shape[0]):
#         for y in range(state.map.shape[1]):
#             if (
#                 not max(
#                     abs(x - state.player_position[0]), abs(y - state.player_position[1])
#                 )
#                 <= 4
#             ):
#                 continue
#             tile = {
#                 "position": {
#                     "x": x - state.player_position[0],
#                     "y": y - state.player_position[1],
#                 },
#                 "visible": max(
#                     abs(x - state.player_position[0]), abs(y - state.player_position[1])
#                 )
#                 <= 4,
#                 "block": BlockType(state.map[x, y]).name.lower(),
#             }
#             if state.mob_map[x, y].max() > 0.5:
#                 tile["mob"] = mob_id_to_name(state.mob_map[x, y].argmax())
#             map_data.append(tile)

#     inventory_data = {
#         "resources": {
#             "wood": int(state.inventory.wood),
#             "stone": int(state.inventory.stone),
#             "coal": int(state.inventory.coal),
#             "iron": int(state.inventory.iron),
#             "diamond": int(state.inventory.diamond),
#             "sapling": int(state.inventory.sapling),
#         },
#         "tools": {},
#     }

#     inventory_data["tools"]["pickaxe"] = {
#         "wood": int(state.inventory.wood_pickaxe),
#         "stone": int(state.inventory.stone_pickaxe),
#         "iron": int(state.inventory.iron_pickaxe),
#     }

#     inventory_data["tools"]["sword"] = {
#         "wood": int(state.inventory.wood_sword),
#         "stone": int(state.inventory.stone_sword),
#         "iron": int(state.inventory.iron_sword),
#     }

#     player_data = {
#         "health": int(state.player_health),
#         "food": int(state.player_food),
#         "drink": int(state.player_drink),
#         "energy": int(state.player_energy),
#         "direction_facing": Action(state.player_direction).name.lower(),
#     }

#     environment_data = {
#         "light_level": float(state.light_level),
#         "is_sleeping": bool(state.is_sleeping),
#         "floor": 0,  # Hardcode to 0 since player_level isn't in EnvState
#     }

#     def to_json_friendly(data):
#         if isinstance(data, dict):
#             return {key: to_json_friendly(value) for key, value in data.items()}
#         elif isinstance(data, (list, tuple)):
#             return [to_json_friendly(item) for item in data]
#         elif isinstance(data, (jnp.ndarray, np.ndarray)):
#             if data.ndim == 0:
#                 return data.item()
#             return [to_json_friendly(item) for item in data]
#         elif isinstance(data, (jnp.bool_, np.bool_)):
#             return bool(data)
#         elif isinstance(data, (jnp.integer, np.integer)):
#             return int(data)
#         elif isinstance(data, (jnp.floating, np.floating)):
#             return float(data)
#         else:
#             return data

#     craftax_state = CraftaxClassicState.create(
#         map=to_json_friendly(map_data),
#         inventory=to_json_friendly(inventory_data),
#         player=to_json_friendly(player_data),
#         environment=to_json_friendly(environment_data),
#         enable_recording=enable_recording,
#     )

#     # Record frame if recording is enabled
#     craftax_state.record_if_enabled(state)

#     return craftax_state


def render_craftax_text(state):
    """Renders the CraftAX Classic state as text representation."""
    obs_dim_array = jnp.array([OBS_DIM[0], OBS_DIM[1]], dtype=jnp.int32)

    # Map
    padded_grid = jnp.pad(
        state.map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=BlockType.OUT_OF_BOUNDS.value,
    )

    tl_corner = state.player_position - obs_dim_array // 2 + MAX_OBS_DIM + 2
    map_view = jax.lax.dynamic_slice(padded_grid, tl_corner, OBS_DIM)

    # Convert blocks to text representation
    block_chars = {
        BlockType.INVALID.value: ".",
        BlockType.OUT_OF_BOUNDS.value: "#",
        BlockType.GRASS.value: "G",
        BlockType.WATER.value: "~",
        BlockType.STONE.value: "S",
        BlockType.TREE.value: "T",
        BlockType.WOOD.value: "W",
        BlockType.PATH.value: "P",
        BlockType.COAL.value: "C",
        BlockType.IRON.value: "I",
        BlockType.DIAMOND.value: "D",
        BlockType.CRAFTING_TABLE.value: "B",  # B for crafting Bench
        BlockType.FURNACE.value: "F",
        BlockType.SAND.value: "_",
        BlockType.LAVA.value: "L",
        BlockType.PLANT.value: "p",
        BlockType.RIPE_PLANT.value: "r",
        BlockType.WALL.value: "X",
        BlockType.DARKNESS.value: " ",
        BlockType.WALL_MOSS.value: "M",
        BlockType.STALAGMITE.value: "^",
        BlockType.SAPPHIRE.value: "$",
        BlockType.RUBY.value: "R",
        BlockType.CHEST.value: "H",
        BlockType.FOUNTAIN.value: "O",
        BlockType.FIRE_GRASS.value: "f",
        BlockType.ICE_GRASS.value: "i",
        BlockType.GRAVEL.value: "g",
        BlockType.FIRE_TREE.value: "t",
        BlockType.ICE_SHRUB.value: "s",
        BlockType.ENCHANTMENT_TABLE_FIRE.value: "E",
        BlockType.ENCHANTMENT_TABLE_ICE.value: "e",
        BlockType.NECROMANCER.value: "N",
        BlockType.GRAVE.value: "v",
        BlockType.GRAVE2.value: "V",
        BlockType.GRAVE3.value: "u",
        BlockType.NECROMANCER_VULNERABLE.value: "n",
    }

    # Create text grid - swap x and y to match game coordinates
    height, width = map_view.shape
    text_grid = []
    for y in range(height):
        row = []
        for x in range(width):
            block_value = int(map_view[y, x])
            row.append(block_chars.get(block_value, "?"))
        text_grid.append(row)

    # Add player at center
    center_y, center_x = obs_dim_array // 2

    # print(state.player_direction)
    # print("Player chars:")
    # print(player_chars)
    player_direction = int(state.player_direction)
    text_grid[center_y][center_x] = player_chars.get(player_direction, "P")
    # print("Final player char:")
    # print(text_grid[center_y][center_x])

    # Add mobs - ensure coordinates are properly aligned
    def add_mob_to_grid(mob_positions, mob_masks, symbol):
        for pos, mask in zip(mob_positions, mob_masks):
            if not mask:
                continue
            # Calculate local position relative to player
            local_pos = pos - state.player_position + obs_dim_array // 2
            # Check if mob is within visible area
            if (local_pos >= 0).all() and (local_pos < obs_dim_array).all():
                y, x = int(local_pos[0]), int(local_pos[1])
                text_grid[y][x] = symbol

    add_mob_to_grid(state.zombies.position, state.zombies.mask, "Z")
    add_mob_to_grid(state.cows.position, state.cows.mask, "c")
    add_mob_to_grid(state.skeletons.position, state.skeletons.mask, "K")
    add_mob_to_grid(state.arrows.position, state.arrows.mask, "a")

    # Convert to string - join rows in correct order
    text_map = "\n".join("".join(row) for row in text_grid)

    return text_map


def render_craftax_classic_text_custom(
    state: EnvState, enable_recording=False, vantage=3
) -> CraftaxClassicState:
    #print("Custom")
    map_data = []
    for x in range(state.map.shape[0]):
        for y in range(state.map.shape[1]):
            if (
                not max(
                    abs(x - state.player_position[0]), abs(y - state.player_position[1])
                )
                <= 4
            ):
                continue
            tile = {
                "position": {
                    "x": x - state.player_position[0],
                    "y": y - state.player_position[1],
                },
                "visible": max(
                    abs(x - state.player_position[0]), abs(y - state.player_position[1])
                )
                <= 4,
                "block": BlockType(state.map[x, y]).name.lower(),
            }
            if state.mob_map[x, y].max() > 0.5:
                tile["mob"] = mob_id_to_name(state.mob_map[x, y].argmax())
            map_data.append(tile)
    text_map = render_craftax_text(state)

    inventory_data = {
        "resources": {
            "wood": int(state.inventory.wood),
            "stone": int(state.inventory.stone),
            "coal": int(state.inventory.coal),
            "iron": int(state.inventory.iron),
            "diamond": int(state.inventory.diamond),
            "sapling": int(state.inventory.sapling),
        },
        "tools": {},
    }

    inventory_data["tools"]["pickaxe"] = {
        "wood": int(state.inventory.wood_pickaxe),
        "stone": int(state.inventory.stone_pickaxe),
        "iron": int(state.inventory.iron_pickaxe),
    }

    inventory_data["tools"]["sword"] = {
        "wood": int(state.inventory.wood_sword),
        "stone": int(state.inventory.stone_sword),
        "iron": int(state.inventory.iron_sword),
    }

    player_data = {
        "health": int(state.player_health),
        "food": int(state.player_food),
        "drink": int(state.player_drink),
        "energy": int(state.player_energy),
        "direction_facing": Action(state.player_direction).name.lower(),
    }

    environment_data = {
        "light_level": float(state.light_level),
        "is_sleeping": bool(state.is_sleeping),
        "floor": 0,  # Hardcode to 0 since player_level isn't in EnvState
    }

    def to_json_friendly(data):
        if isinstance(data, dict):
            return {key: to_json_friendly(value) for key, value in data.items()}
        elif isinstance(data, (list, tuple)):
            return [to_json_friendly(item) for item in data]
        elif isinstance(data, (jnp.ndarray, np.ndarray)):
            if data.ndim == 0:
                return data.item()
            return [to_json_friendly(item) for item in data]
        elif isinstance(data, (jnp.bool_, np.bool_)):
            return bool(data)
        elif isinstance(data, (jnp.integer, np.integer)):
            return int(data)
        elif isinstance(data, (jnp.floating, np.floating)):
            return float(data)
        else:
            return data

    pixels = render_craftax_pixels(state, BLOCK_PIXEL_SIZE_HUMAN)
    frame = np.array(pixels).astype(np.uint8)
    craftax_state = CraftaxClassicState(
        map_full=to_json_friendly(map_data),
        map_compact=text_map,
        inventory=to_json_friendly(inventory_data),
        player=to_json_friendly(player_data),
        environment=to_json_friendly(environment_data),
        image=frame,
        recorder=None if not enable_recording else CraftaxRecorder(),
        vantage=vantage,
    )

    # Record frame if recording is enabled
    if enable_recording:
        craftax_state.recorder.record_frame(state)

    return craftax_state


class CraftaxRecorder:
    def __init__(self, save_dir="recordings"):
        self.save_dir = Path(save_dir)
        self.save_dir.mkdir(parents=True, exist_ok=True)
        self.frames = []

    def record_frame(self, state):
        pixels = render_craftax_pixels(state, BLOCK_PIXEL_SIZE_HUMAN)
        # Convert from JAX array to numpy and ensure uint8 format
        frame = np.array(pixels).astype(np.uint8)
        self.frames.append(frame)
        # print("Added frame - total frames:", len(self.frames))

    def save_video(self, sfilename, fps=1):
        save_path = self.save_dir / filename
        imageio.mimsave(str(save_path), self.frames, fps=fps, codec="mpeg4")
        # print(f"Saved video to {save_path} - {len(self.frames)} frames")
        # Clear frames after saving
        self.frames = []



from dataclasses import dataclass
from typing import Dict, List

import jax
from craftax.craftax.constants import *
from craftax.craftax.craftax_state import EnvState
from craftax.craftax.util.game_logic_utils import is_boss_vulnerable

from craftaxlm.full.metadata import (
    CRAFTAX_FULL_BACKDROP_BLOCK_TYPES,
    CRAFTAX_FULL_HIGH_SALIENCE_ITEMS,
    CRAFTAX_FULL_HIGH_SALIENCE_MOBS,
    CRAFTAX_FULL_HIGH_SALIENCE_OBJECTS,
    CRAFTAX_FULL_LOW_SALIENCE_ITEMS,
    CRAFTAX_FULL_LOW_SALIENCE_MOBS,
    CRAFTAX_FULL_LOW_SALIENCE_OBJECTS,
)
from craftaxlm.shared import (
    CraftaxState,
    get_armour_level,
    level_to_enchantment,
    level_to_material,
    mob_id_to_name,
)


@dataclass
class CraftaxFullState(CraftaxState):
    map: List[Dict]
    inventory: Dict
    player: Dict
    environment: Dict

    def render_map_to_text(self, ignore_distant_low_salience=True):
        unique_blocks = list(
            set([tile["block"] for tile in self.map if "block" in tile])
        )
        if not set(unique_blocks).issubset(
            set(
                CRAFTAX_FULL_BACKDROP_BLOCK_TYPES
                + CRAFTAX_FULL_LOW_SALIENCE_OBJECTS
                + CRAFTAX_FULL_HIGH_SALIENCE_OBJECTS
            )
        ):
            raise ValueError(
                f"Unknown block types: {set(unique_blocks) - set(CRAFTAX_FULL_BACKDROP_BLOCK_TYPES + CRAFTAX_FULL_LOW_SALIENCE_OBJECTS + CRAFTAX_FULL_HIGH_SALIENCE_OBJECTS)}"
            )
        unique_mobs = list(set([tile["mob"] for tile in self.map if "mob" in tile]))
        if not set(unique_mobs).issubset(
            set(CRAFTAX_FULL_LOW_SALIENCE_MOBS + CRAFTAX_FULL_HIGH_SALIENCE_MOBS)
        ):
            raise ValueError(
                f"Unknown mob types: {set(unique_mobs) - set(CRAFTAX_FULL_LOW_SALIENCE_MOBS + CRAFTAX_FULL_HIGH_SALIENCE_MOBS)}"
            )
        unique_items = list(set([tile["item"] for tile in self.map if "item" in tile]))
        if not set(unique_items).issubset(
            set(CRAFTAX_FULL_LOW_SALIENCE_ITEMS + CRAFTAX_FULL_HIGH_SALIENCE_ITEMS)
        ):
            raise ValueError(
                f"Unknown item types: {set(unique_items) - set(CRAFTAX_FULL_LOW_SALIENCE_ITEMS + CRAFTAX_FULL_HIGH_SALIENCE_ITEMS)}"
            )

        def count_nearby_blocks(center_x, center_y, radius):
            block_counts = {}
            for tile in self.map:
                dx = abs(tile["position"]["x"] - center_x)
                dy = abs(tile["position"]["y"] - center_y)
                if dx <= radius and dy <= radius:
                    if "block" not in tile:
                        continue
                    block_type = tile["block"]
                    if block_type in CRAFTAX_FULL_BACKDROP_BLOCK_TYPES:
                        block_counts[block_type] = block_counts.get(block_type, 0) + 1
            return block_counts

        center_x, center_y = 0, 0
        nearby_blocks = count_nearby_blocks(center_x, center_y, 2)
        backdrop = (
            max(nearby_blocks, key=nearby_blocks.get) if nearby_blocks else "unknown"
        )

        low_salience_objects = CRAFTAX_FULL_LOW_SALIENCE_OBJECTS.copy()
        low_salience_objects.extend(
            [obj for obj in CRAFTAX_FULL_BACKDROP_BLOCK_TYPES if obj != backdrop]
        )

        def describe_xy(x, y):
            description = ""
            if x > 0:
                description += f"{x} steps down"
            elif x < 0:
                description += f"{-x} steps up"
            if (x != 0) and (y != 0):
                description += " and "
            if y < 0:
                description += f"{-y} steps left"
            elif y > 0:
                description += f"{y} steps right"
            return description

        periphery = []
        for direction in [
            (-1, -1),
            (-1, 0),
            (-1, 1),
            (0, -1),
            (0, 1),
            (1, -1),
            (1, 0),
            (1, 1),
        ]:
            for distance in range(1, 6):
                x = center_x + direction[0] * distance
                y = center_y + direction[1] * distance
                for tile in self.map:
                    if not tile["visible"]:
                        continue
                    if tile["position"]["x"] == x and tile["position"]["y"] == y:
                        if tile["block"] in low_salience_objects:
                            periphery.append(
                                f"{tile['block'].capitalize()} is {describe_xy(x, y)}"
                            )
                            if ignore_distant_low_salience:
                                break
                        if (
                            "mob" in tile
                            and tile["mob"] in CRAFTAX_FULL_LOW_SALIENCE_MOBS
                        ):
                            periphery.append(f"A {tile['mob']} is {describe_xy(x, y)}")
                            if ignore_distant_low_salience:
                                break
                        if (
                            "item" in tile
                            and tile["item"] in CRAFTAX_FULL_LOW_SALIENCE_ITEMS
                        ):
                            periphery.append(f"A {tile['item']} is {describe_xy(x, y)}")
                            if ignore_distant_low_salience:
                                break

        high_salience = []
        for tile in self.map:
            if not tile["visible"]:
                continue
            if "mob" in tile and tile["mob"] in CRAFTAX_FULL_HIGH_SALIENCE_MOBS:
                high_salience.append(
                    f"A {tile['mob']} is {describe_xy(tile['position']['x'], tile['position']['y'])}"
                )
            if "item" in tile and tile["item"] in CRAFTAX_FULL_HIGH_SALIENCE_ITEMS:
                high_salience.append(
                    f"A {tile['item']} is {describe_xy(tile['position']['x'], tile['position']['y'])}"
                )
            if tile["block"] in CRAFTAX_FULL_HIGH_SALIENCE_OBJECTS:
                high_salience.append(
                    f"{tile['block'].capitalize()} is {describe_xy(tile['position']['x'], tile['position']['y'])}"
                )

        facing_and_position = {
            "up": "is one steps up",
            "down": "is one steps down",
            "left": "is one steps left",
            "right": "is one steps right",
        }
        return {
            "terrain_underneath_you": backdrop,
            "surroundings": periphery + high_salience,
            "object_you_are_facing": (
                [
                    surrounding_object.split(" is")[0]
                    for surrounding_object in periphery + high_salience
                    if facing_and_position[self.player["direction"]]
                    in surrounding_object
                ]
                + ["No object directly in front of you"]
            )[0],
        }

    def render_inventory_to_text(self, include_absent_inventory=True):
        def process_inventory(inv_dict):
            processed = {}
            for key, value in inv_dict.items():
                if isinstance(value, dict):
                    processed[key] = process_inventory(value)
                elif isinstance(value, bool) and (include_absent_inventory or value):
                    processed[key] = value
                elif include_absent_inventory or value:
                    processed[key] = value is not None
            return processed

        return process_inventory(self.inventory)

    def render_environment_to_text(self, include_absent_environment_attributes=True):
        attributes_to_hide = [
            "learned_fireball",
            "learned_iceball",
            "floor",
            "ladder_open",
            "is_boss_vulnerable",
        ]
        environment = {}
        for key, value in self.environment.items():
            if key not in attributes_to_hide and (
                include_absent_environment_attributes or value
            ):
                environment[key] = value
        return environment


def render_craftax_text_custom(state: EnvState) -> CraftaxFullState:
    obs_dim_array = jnp.array([OBS_DIM[0], OBS_DIM[1]], dtype=jnp.int32)
    map = state.map[state.player_level]
    padded_grid = jnp.pad(
        map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=BlockType.OUT_OF_BOUNDS.value,
    )
    tl_corner = state.player_position - obs_dim_array // 2 + MAX_OBS_DIM + 2
    map_view = jax.lax.dynamic_slice(padded_grid, tl_corner, OBS_DIM)

    padded_items_map = jnp.pad(
        state.item_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=ItemType.NONE.value,
    )
    item_map_view = jax.lax.dynamic_slice(padded_items_map, tl_corner, OBS_DIM)

    mob_types_per_class = 8
    mob_map = jnp.zeros((*OBS_DIM, 5 * mob_types_per_class), dtype=jnp.int32)

    def _add_mob_to_map(carry, mob_index):
        mob_map, mobs, mob_class_index = carry
        local_position = (
            mobs.position[mob_index]
            - state.player_position
            + jnp.array([OBS_DIM[0], OBS_DIM[1]]) // 2
        )
        on_screen = jnp.logical_and(
            local_position >= 0, local_position < jnp.array([OBS_DIM[0], OBS_DIM[1]])
        ).all()
        on_screen *= mobs.mask[mob_index]
        mob_identifier = mob_class_index * mob_types_per_class + mobs.type_id[mob_index]
        mob_map = mob_map.at[local_position[0], local_position[1], mob_identifier].set(
            on_screen.astype(jnp.int32)
        )
        return (mob_map, mobs, mob_class_index), None

    for mob_type, mob_class_index in [
        (state.melee_mobs, 0),
        (state.passive_mobs, 1),
        (state.ranged_mobs, 2),
        (state.mob_projectiles, 3),
        (state.player_projectiles, 4),
    ]:
        (mob_map, _, _), _ = jax.lax.scan(
            _add_mob_to_map,
            (
                mob_map,
                jax.tree_util.tree_map(lambda x: x[state.player_level], mob_type),
                mob_class_index,
            ),
            jnp.arange(mob_type.mask.shape[1]),
        )

    padded_light_map = jnp.pad(
        state.light_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=0.0,
    )
    light_map_view = jax.lax.dynamic_slice(padded_light_map, tl_corner, OBS_DIM) > 0.05

    map_data = []
    for x in range(OBS_DIM[0]):
        for y in range(OBS_DIM[1]):
            tile = {
                "position": {"x": y - OBS_DIM[1] // 2, "y": x - OBS_DIM[0] // 2},
                "visible": bool(light_map_view[x, y]),
            }
            if light_map_view[x, y]:
                if mob_map[x, y].max() > 0.5:
                    tile["mob"] = mob_id_to_name(mob_map[x, y].argmax())
                if item_map_view[x, y] != ItemType.NONE.value:
                    tile["item"] = ItemType(item_map_view[x, y]).name.lower()
                tile["block"] = BlockType(map_view[x, y]).name.lower()
            map_data.append(tile)
    inventory_data = {
        "resources": {
            "wood": state.inventory.wood,
            "stone": state.inventory.stone,
            "coal": state.inventory.coal,
            "iron": state.inventory.iron,
            "diamond": state.inventory.diamond,
            "sapphire": state.inventory.sapphire,
            "ruby": state.inventory.ruby,
            "sapling": state.inventory.sapling,
            "torch": state.inventory.torches,
            "arrow": state.inventory.arrows,
            "book": state.inventory.books,
        },
        "tools": {},
        "potions": {
            "red": state.inventory.potions[0],
            "green": state.inventory.potions[1],
            "blue": state.inventory.potions[2],
            "pink": state.inventory.potions[3],
            "cyan": state.inventory.potions[4],
            "yellow": state.inventory.potions[5],
        },
        "armor": {},
    }

    if state.inventory.pickaxe > 0:
        inventory_data["tools"]["pickaxe"] = {
            "material": level_to_material(state.inventory.pickaxe)
        }
    if state.inventory.sword > 0:
        inventory_data["tools"]["sword"] = {
            "material": level_to_material(state.inventory.sword),
            "enchantment": level_to_enchantment(state.sword_enchantment),
        }
    if state.inventory.bow > 0:
        inventory_data["tools"]["bow"] = {
            "enchantment": level_to_enchantment(state.bow_enchantment)
        }

    for i, piece in enumerate(["helmet", "chestplate", "leggings", "boots"]):
        if state.inventory.armour[i] > 0:
            inventory_data["armor"][piece] = {
                "material": get_armour_level(state.inventory.armour[i]),
                "enchantment": level_to_enchantment(state.armour_enchantments[i]),
            }
    player_data = {
        "health": state.player_health,
        "food": state.player_food,
        "drink": state.player_drink,
        "energy": state.player_energy,
        "mana": state.player_mana,
        "xp": state.player_xp,
        "dexterity": state.player_dexterity,
        "strength": state.player_strength,
        "intelligence": state.player_intelligence,
        "direction": Action(state.player_direction).name.lower(),
    }

    environment_data = {
        "light_level": state.light_level,
        "is_sleeping": state.is_sleeping,
        "is_resting": state.is_resting,
        "learned_fireball": state.learned_spells[0],
        "learned_iceball": state.learned_spells[1],
        "floor": state.player_level,
        "ladder_open": state.monsters_killed[state.player_level]
        >= MONSTERS_KILLED_TO_CLEAR_LEVEL,
        "is_boss_vulnerable": is_boss_vulnerable(state),
    }

    def to_json_friendly(data):
        if isinstance(data, dict):
            return {key: to_json_friendly(value) for key, value in data.items()}
        elif isinstance(data, (list, tuple)):
            return [to_json_friendly(item) for item in data]
        elif isinstance(data, (jnp.ndarray, np.ndarray)):
            if data.ndim == 0:
                return data.item()
            return [to_json_friendly(item) for item in data]
        elif isinstance(data, (jnp.bool_, np.bool_)):
            return bool(data)
        elif isinstance(data, (jnp.integer, np.integer)):
            return int(data)
        elif isinstance(data, (jnp.floating, np.floating)):
            return float(data)
        else:
            return data

    return CraftaxFullState(
        map=to_json_friendly(map_data),
        inventory=to_json_friendly(inventory_data),
        player=to_json_friendly(player_data),
        environment=to_json_friendly(environment_data),
    )



from pathlib import Path
import numpy as np
import imageio
from typing import List, Optional
#ADD full craftax recording later
from craftax.craftax_classic.renderer import render_craftax_pixels
from craftax.craftax.craftax_state import EnvState
from craftax.craftax.constants import BLOCK_PIXEL_SIZE_HUMAN


class EpisodeRecorder:
    """Records frames from a Craftax episode."""

    def __init__(self, save_dir: str = "recordings", enabled: bool = True, ):
        self.save_dir = Path(save_dir)
        self.enabled = enabled
        if enabled:
            self.save_dir.mkdir(parents=True, exist_ok=True)
            self.frames: List[np.ndarray] = []

    def record_frame(self, state: EnvState) -> None:
        """Record a single frame from the environment state."""
        if not self.enabled:
            return

        pixels = render_craftax_pixels(state, BLOCK_PIXEL_SIZE_HUMAN)
        frame = np.array(pixels).astype(np.uint8)
        self.frames.append(frame)

    def save_video(self, filename: str, fps: int = 10) -> Optional[Path]:
        """Save recorded frames as a video file."""
        if not self.enabled or not self.frames:
            return None

        save_path = self.save_dir / filename
        imageio.mimsave(str(save_path), self.frames, fps=fps, codec="mpeg4")
        #print(f"Saved {len(self.frames)} frames to {save_path}")
        self.frames = []  # Clear frames after saving
        return save_path

    def clear(self) -> None:
        """Clear recorded frames without saving."""
        if self.enabled:
            self.frames = []



from setuptools import setup, find_packages

setup(
    name="craftaxlm",
    version="0.0.29",
    packages=find_packages(),
    install_requires=[
        "absl-py==2.1.0",
        "backports.tarfile==1.2.0",
        "black==24.4.2",
        "build==1.2.1",
        "certifi==2024.7.4",
        "cfgv==3.4.0",
        "charset-normalizer==3.3.2",
        "chex==0.1.86",
        "click==8.1.7",
        "cloudpickle==3.0.0",
        "contourpy==1.2.1",
        "craftax==1.4.3",
        "cycler==0.12.1",
        "decorator==5.1.1",
        "distlib==0.3.8",
        "distrax==0.1.5",
        "dm-tree==0.1.8",
        "docker-pycreds==0.4.0",
        "docutils==0.21.2",
        "etils==1.9.2",
        "Farama-Notifications==0.0.4",
        "filelock==3.15.4",
        "flax==0.8.5",
        "fonttools==4.53.1",
        "fsspec==2024.6.1",
        "gast==0.6.0",
        "gitdb==4.0.11",
        "GitPython==3.1.43",
        "gym==0.26.2",
        "gym-notices==0.0.8",
        "gymnasium==0.29.1",
        "gymnax==0.0.8",
        "identify==2.6.0",
        "idna==3.7",
        "imageio==2.34.2",
        "importlib_metadata==8.2.0",
        "importlib_resources==6.4.0",
        "jaraco.classes==3.4.0",
        "jaraco.context==5.3.0",
        "jaraco.functools==4.0.2",
        "jax==0.4.30",
        "jaxlib==0.4.30",
        "keyring==25.3.0",
        "kiwisolver==1.4.5",
        "markdown-it-py==3.0.0",
        "matplotlib==3.9.1",
        "mdurl==0.1.2",
        "ml-dtypes==0.4.0",
        "more-itertools==10.4.0",
        "msgpack==1.0.8",
        "mypy-extensions==1.0.0",
        "nest-asyncio==1.6.0",
        "nh3==0.2.18",
        "nodeenv==1.9.1",
        "numpy==2.0.1",
        "opt-einsum==3.3.0",
        "optax==0.2.3",
        "orbax-checkpoint==0.5.23",
        "packaging==24.1",
        "pandas==2.2.2",
        "pathspec==0.12.1",
        "pillow==10.4.0",
        "pkginfo==1.10.0",
        "platformdirs==4.2.2",
        "pre-commit==3.8.0",
        "protobuf==5.27.2",
        "psutil==6.0.0",
        "pygame==2.6.0",
        "Pygments==2.18.0",
        "pyparsing==3.1.2",
        "pyproject_hooks==1.1.0",
        "python-dateutil==2.9.0.post0",
        "pytz==2024.1",
        "PyYAML==6.0.1",
        "readme_renderer==44.0",
        "requests==2.32.3",
        "requests-toolbelt==1.0.0",
        "rfc3986==2.0.0",
        "rich==13.7.1",
        "scipy==1.14.0",
        "seaborn==0.13.2",
        "sentry-sdk==2.11.0",
        "setproctitle==1.3.3",
        "six==1.16.0",
        "smmap==5.0.1",
        "tensorflow-probability==0.24.0",
        "tensorstore==0.1.63",
        "toolz==0.12.1",
        "twine==5.1.1",
        "typing_extensions==4.12.2",
        "tzdata==2024.1",
        "urllib3==2.2.2",
        "virtualenv==20.26.3",
        "wandb==0.17.5",
        "zipp==3.19.2",
    ],
    author="Josh Purtell",
    author_email="jmvpurtell@gmail.com",
    description="Craftax LM: Benchmarking LM Agents with Craftax",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JoshuaPurtell/craftaxlm",
    license="MIT",
)



import math
from abc import abstractclassmethod
from dataclasses import dataclass
from typing import Dict, List, Literal, Tuple

import jax


def mob_id_to_name(id):
    if id == 0:
        return "Zombie"
    elif id == 1:
        return "Gnome Warrior"
    elif id == 2:
        return "Orc Soldier"
    elif id == 3:
        return "Lizard"
    elif id == 4:
        return "Knight"
    elif id == 5:
        return "Troll"
    elif id == 6:
        return "Pigman"
    elif id == 7:
        return "Frost Troll"
    elif id == 8:
        return "Cow"
    elif id == 9:
        return "Bat"
    elif id == 10:
        return "Snail"
    elif id == 16:
        return "Skeleton"
    elif id == 17:
        return "Gnome Archer"
    elif id == 18:
        return "Orc Mage"
    elif id == 19:
        return "Kobold"
    elif id == 20:
        return "Archer"
    elif id == 21:
        return "Deep Thing"
    elif id == 22:
        return "Fire Elemental"
    elif id == 23:
        return "Ice Elemental"
    elif id == 24:
        return "Arrow"
    elif id == 25:
        return "Dagger"
    elif id == 26:
        return "Fireball"
    elif id == 27:
        return "Iceball"
    elif id == 28:
        return "Arrow"
    elif id == 29:
        return "Slimeball"
    elif id == 30:
        return "Fireball"
    elif id == 31:
        return "Iceball"
    elif id == 32:
        return "Arrow (Player)"
    elif id == 33:
        return "Dagger (Player)"
    elif id == 34:
        return "Fireball (Player)"
    elif id == 35:
        return "Iceball (Player)"
    elif id == 36:
        return "Arrow (Player)"
    elif id == 37:
        return "Slimeball (Player)"
    elif id == 38:
        return "Fireball (Player)"
    elif id == 39:
        return "Iceball (Player)"


def level_to_material(level):
    if level == 1:
        return "Wood"
    elif level == 2:
        return "Stone"
    elif level == 3:
        return "Iron"
    elif level == 4:
        return "Diamond"


def level_to_enchantment(level):
    if level == 0:
        return "No"
    if level == 1:
        return "Fire"
    elif level == 2:
        return "Ice"


def get_armour_level(level):
    if level == 1:
        return "Iron"
    elif level == 2:
        return "Diamond"


@dataclass
class CraftaxState:
    map_full: List[Dict]
    map_compact: str
    inventory: Dict
    player: Dict
    environment: Dict

    @abstractclassmethod
    def render_map_to_text(self, ignore_distant_low_salience=True):
        pass

    @abstractclassmethod
    def render_inventory_to_text(self, include_absent_inventory=True):
        pass

    @abstractclassmethod
    def render_environment_to_text(self, include_absent_environment_attributes=True):
        pass

    def render_json_to_text_via_md(self, json_data: Dict) -> str:
        keys_to_not_format = [
            "resources",
            "tools",
            "potions",
            "armor",
            "player",
            "environment",
        ]

        def format_as_list(data, level):
            list_output = ""
            if isinstance(data, dict):
                for k, v in data.items():
                    if isinstance(v, dict):
                        list_output += (
                            f"{' ' * level}- {k}:\n{format_as_list(v, level + 2)}"
                        )
                    else:
                        list_output += f"{' ' * level}- {k}: {v}\n"
            elif isinstance(data, list):
                for item in data:
                    list_output += f"{' ' * level}- {item}\n"
            else:
                list_output += f"{' ' * level}- {data}\n"
            return list_output

        def dict_to_md(data, level=1, parent_key=""):
            md_output = ""
            for key, value in data.items():
                full_key = f"{parent_key}.{key}" if parent_key else key
                md_output += f"{'#' * level} {key.capitalize()}\n"
                if key in keys_to_not_format:
                    md_output += format_as_list(value, level)
                elif isinstance(value, dict):
                    md_output += dict_to_md(value, level + 1, full_key)
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, dict):
                            md_output += dict_to_md(item, level + 1, full_key)
                        else:
                            md_output += f"- {item}\n"
                else:
                    md_output += f"{value}\n"
            return md_output

        return dict_to_md(json_data).strip()

    def render_json_to_text_via_xml(self, json_data: Dict) -> str:
        def dict_to_xml(data):
            xml_output = ""
            for key, value in data.items():
                xml_output += f"<{key}>"
                if isinstance(value, dict):
                    xml_output += dict_to_xml(value)
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, dict):
                            xml_output += dict_to_xml(item)
                        else:
                            xml_output += str(item)
                else:
                    xml_output += str(value)
                xml_output += f"</{key}>"
            return xml_output

        return dict_to_xml(json_data).strip()

    def render_to_text_simple(
        self,
        verbose=True,
        formatting: Literal["md", "xml"] = "md",
        map_format: Literal["full", "compact"] = "full",
    ) -> str:
        if map_format == "full":
            rendered_map = self.render_map_to_text(
                ignore_distant_low_salience=not verbose
            )
        else:
            rendered_map = self.map_compact

        rendered_inventory = self.render_inventory_to_text(
            include_absent_inventory=verbose
        )
        rendered_environment = self.render_environment_to_text(
            include_absent_environment_attributes=verbose
        )
        if formatting == "md":
            return self.render_json_to_text_via_md(
                {
                    "map": rendered_map,
                    "inventory": rendered_inventory,
                    "player": self.player,
                    "environment": rendered_environment,
                }
            )
        elif formatting == "xml":
            return self.render_json_to_text_via_xml(
                {
                    "map": rendered_map,
                    "inventory": rendered_inventory,
                    "player": self.player,
                    "environment": rendered_environment,
                }
            )
        else:
            raise ValueError(f"Unknown formatting: {formatting}")


class CraftaxBaseACI:
    def __init__(
        self,
        seed=0,
        actions_to_start_with: List[int] = [],
        verbose=True,
        formatting: Literal["md", "xml"] = "md",
        map_format: Literal["full", "compact"] = "full",
        vantage: int = 3,
    ):
        self.formatting = formatting
        self.map_format = map_format
        self.verbose = verbose
        self.vantage = vantage
        rng = jax.random.PRNGKey(seed)
        rng, _rng = jax.random.split(rng)
        self.rngs = jax.random.split(_rng, 3)
        self.reset()
        if actions_to_start_with:
            self.go_forward(actions_to_start_with)

    def reset(self):
        self.env = self.make_env()
        self.env_params = self.env.default_params
        obs, self.state = self.env.reset(self.rngs[0], self.env_params)
        self.starting_obs = self.create_starting_obs()
        self.action_history = []
        self.achievements = {}
        self.achievement_deltas = []

    def make_env(self):
        pass

    def create_starting_obs(self):
        pass

    def go_back(self, n_steps):
        actions = self.action_history[-n_steps:]
        self.reset()
        return self.go_forward(actions)

    def go_forward(self, actions):
        step_info = {}
        for action in actions:
            step_info = self._step(action)
        return step_info

    def render_achivements(self, info):
        achievements = {}
        for key, value in info.items():
            if key.startswith("Achievements/"):
                achievement_name = key.split("/")[-1]
                achievements[achievement_name] = float(value)
        return {
            "achievements": achievements,
            "discount": float(info.get("discount", 1.0)),
        }

    def get_achievement_delta(self, achievements):
        delta = []
        for key, value in achievements["achievements"].items():
            if key in self.achievements:
                if value > self.achievements[key]:
                    delta.append(key)
        self.achievements = achievements["achievements"]
        return delta

    def map_action_string_to_int(self, action_string: str) -> int:
        pass

    def _step(self, action):
        _, state, reward, done, info = self.env.step(
            self.rngs[2], self.state, action, self.env_params
        )
        achievements = self.get_achievements(state)
        achievement_delta = self.get_achievement_delta(achievements)
        assert achievement_delta is not None
        self.achievement_deltas.append(achievement_delta)
        self.state = state
        step_info = self.create_step_info(state, reward, done)
        return step_info
    
    def check_efficacy(self, state_before, state_after, action):
        pass

    @abstractclassmethod
    def get_achievements(self, state):
        pass

    @abstractclassmethod
    def create_step_info(self, state, reward, done):
        pass

    def multistep(self, actions: List[str]) -> Tuple[List[Dict], List[float], bool]:
        done = False
        step_infos = []
        rewards = []
        for action in actions:
            step_info = self._step(self.map_action_string_to_int(action))
            step_infos.append(step_info)
            rewards.append(step_info["reward"])
            done = step_info["done"]
            if done:
                break
        return step_infos, rewards, done

    def terminate(self):
        return self.achievements


